from ..core.project_manager import *
from uquake.nlloc.nlloc import *
from uquake.core.event import (Catalog, Event, RayCollection, Pick,
                               CreationInfo, Ray)
from uquake.core import UTCDateTime
import numpy as np


def calculate_uncertainty(point_cloud):
    """
    :param point_cloud: location point cloud
    :return: obspy.core.event.OriginUncertainty()
    """

    v, u = np.linalg.eig(np.cov(point_cloud.T))

    major_axis_index = np.argmax(v)

    uncertainty = np.sort(np.sqrt(v))[-1::-1]

    h = np.linalg.norm(u[major_axis_index, :-1])
    vert = u[major_axis_index, -1]

    major_axis_plunge = np.arctan2(-vert, h)
    x = u[major_axis_index, 0]
    y = u[major_axis_index, 1]
    major_axis_azimuth = np.arctan2(x, y)
    major_axis_rotation = 0

    ce = ConfidenceEllipsoid(semi_major_axis_length=uncertainty[0],
                             semi_intermediate_axis_length=uncertainty[1],
                             semi_minor_axis_length=uncertainty[2],
                             major_axis_plunge=major_axis_plunge,
                             major_axis_azimuth=major_axis_azimuth,
                             major_axis_rotation=major_axis_rotation)

    return OriginUncertainty(confidence_ellipsoid=ce,
                             preferred_description='confidence ellipsoid',
                             confidence_level=68)


class NLLOCResult(object):
    def __init__(self, hypocenter: np.array, event_time: UTCDateTime,
                 scatter_cloud: np.ndarray, rays: list,
                 observations: Observations, evaluation_mode: str,
                 evaluation_status: str):
        self.hypocenter = hypocenter
        self.event_time = event_time
        self.scatter_cloud = scatter_cloud
        self.rays = rays
        self.observations = observations
        self.evaluation_mode = evaluation_mode
        self.evaluation_status = evaluation_status

        self.uncertainty_ellipsoid = calculate_uncertainty(
            self.scatter_cloud[:, :-1])

        self.creation_info = CreationInfo(author='uQuake-nlloc',
                                          creation_time=UTCDateTime.now())

    def __repr__(self):
        ce = self.uncertainty_ellipsoid.confidence_ellipsoid
        out_str = f"""
        time (UTC)  : {self.t}
        location    : x- {self.x:>10.1f} (m)
                      y- {self.y:>10.1f} (m)
                      z- {self.z:>10.1f} (m)
        uncertainty : {ce.semi_major_axis_length:0.1f} (1 std - m)
        """
        return out_str

    def __add__(self, other: Event):
        if not (isinstance(other, Catalog) | isinstance(other, Event)):
            raise TypeError(f'object type {type(Event)} or'
                            f'{type(Catalog)} expected. Object of type '
                            f'{type(other)} provided')
        return self.append_to_event(other)

    @property
    def loc(self):
        return self.hypocenter

    @property
    def x(self):
        return self.loc[0]

    @property
    def y(self):
        return self.loc[1]

    @property
    def z(self):
        return self.loc[2]

    @property
    def t(self):
        return self.event_time

    @property
    def arrivals(self) -> list:
        arrivals = []
        for pick in self.observations.picks:
            travel_time = pick.time - self.t
            phase = pick.phase_hint
            sensor_code = pick.sensor
            for ray in self.rays:
                if (ray.sensor_code == sensor_code) & (ray.phase == phase):
                    break
            distance = ray.length

            time_residual = Arrival.calculate_time_residual(ray.travel_time,
                                                            travel_time)

            time_weight = 1
            azimuth = ray.azimuth
            takeoff_angle = ray.takeoff_angle

            arrival = Arrival(phase=phase, distance=distance,
                              time_residual=time_residual,
                              time_weight=time_weight, azimuth=azimuth,
                              takeoff_angle=takeoff_angle,
                              pick_id=pick.resource_id)

            arrivals.append(arrival)

        return arrivals

    @property
    def origin(self):
        origin = Origin(x=self.x, y=self.y, z=self.z, time=self.t,
                        evaluation_mode=self.evaluation_mode,
                        evaluation_status=self.evaluation_status,
                        epicenter_fixed=False, method_id='uQuake-NLLOC',
                        creation_info=self.creation_info,
                        arrivals=self.arrivals,
                        origin_uncertainty=self.uncertainty_ellipsoid,
                        rays=self.rays)
        return origin

    @property
    def event(self):
        return self.export_as_event()

    def append_to_event(self, event: Event) -> Event:
        o = self.origin
        if isinstance(event, Catalog):
            event[0].append_origin_as_preferred_origin(o)
        elif isinstance(event, Event):
            event.append_origin_as_preferred_origin(o)
        return event

    def export_as_event(self):
        o = self.origin
        e = Event(origins=[o], picks=self.observations.picks)
        e.preferred_origin_id = o.resource_id
        return e


class NLLOC(ProjectManager):
    def __init__(self, path: Path, project_name: str, network_code: str,
                 use_srces: bool=False):

        super().__init__(path, project_name, network_code, use_srces=use_srces)

        self.run_id = str(uuid4())
        self.current_run_directory = self.root_directory / 'run' / self.run_id
        self.current_run_directory.mkdir(parents=True, exist_ok=False)

        self.output_file_path = self.current_run_directory / 'outputs'
        self.output_file_path.mkdir(parents=True, exist_ok=True)

        self.observation_path = self.current_run_directory / 'observations'
        self.observation_path.mkdir(parents=True, exist_ok=True)
        self.observation_file_name = 'observations.obs'
        self.observation_file = self.observation_path / \
                                self.observation_file_name
        self.observations = None

        self.template_directory = self.root_directory / 'templates'

        self.template_directory.mkdir(parents=True, exist_ok=True)

        self.template_ctrl_file = self.template_directory / \
                                  'ctrl_template.pickle'

        if self.template_ctrl_file.exists():
            with open(self.template_ctrl_file, 'rb') as template_ctrl:
                self.control_template = pickle.load(template_ctrl)
        else:
            self.control_template = None
            self.add_template_control()

        self.control_file = self.current_run_directory / 'run.nll'

        self.last_event_hypocenter = None
        self.last_event_time = None

        self.nlloc_settings_file = self.config_location / 'nlloc.toml'

        if not self.settings_file.is_file():
            settings_template = Path(os.path.realpath(__file__)).parent / \
                                    '../settings/nlloc_settings_template.toml'

            shutil.copyfile(settings_template, self.nlloc_settings_file)

        self.settings = Settings(str(self.config_location))

    def add_template_control(self, control=Control(message_flag=1),
                             transformation=GeographicTransformation(),
                             locsig=None, loccom=None,
                             locsearch=LocSearchOctTree.init_default(),
                             locmeth=LocationMethod.init_default(),
                             locgau=GaussianModelErrors.init_default(),
                             locqual2err=LocQual2Err(0.0001, 0.0001, 0.0001,
                                                     0.0001, 0.0001),
                             **kwargs):

        if not isinstance(control, Control):
            raise TypeError(f'control is type {type(control)}. '
                            f'control must be type {Control}.')

        if not issubclass(type(transformation), GeographicTransformation):
            raise TypeError(f'transformation is type {type(transformation)}. '
                            f'expecting type'
                           f'{GeographicTransformation}.')

        if not locsearch.type == 'LOCSEARCH':
            raise TypeError(f'locsearch is type {type(locsearch)}'
                            f'expecting type '
                            f'{LocSearchGrid} '
                            f'{LocSearchMetropolis}, or '
                            f'{LocSearchOctTree}.')

        if not isinstance(locmeth, LocationMethod):
            raise TypeError(f'locmeth is type {type(locmeth)}, '
                            f'expecting type {LocationMethod}')

        if not isinstance(locgau, GaussianModelErrors):
            raise TypeError(f'locgau is type {type(locgau)}, '
                            f'expecting type {GaussianModelErrors}')

        if not isinstance(locqual2err, LocQual2Err):
            raise TypeError(f'locqual2err is type {type(locqual2err)}, '
                            f'expecting type {LocQual2Err}')

        dict_out = {'control': control,
                    'transformation': transformation,
                    'locsig': locsig,
                    'loccom': loccom,
                    'locsearch': locsearch,
                    'locmeth': locmeth,
                    'locgau': locgau,
                    'locqual2err': locqual2err}

        with open(self.template_ctrl_file, 'wb') as template_ctrl:
            pickle.dump(dict_out, template_ctrl)

        self.control_template = dict_out

    def write_control_file(self):
        with open(self.control_file, 'w') as control_file:
            control_file.write(self.control)

    def __add_observations__(self, observations):
        """
        adding observations to the project
        :param observations: Observations
        """
        if not isinstance(observations, Observations):
            raise TypeError(f'observations is type {type(observations)}. '
                            f'observations must be type {Observations}.')
        self.observation_path.mkdir(parents=True, exist_ok=True)
        observations.write(self.observation_file_name,
                           path=self.observation_path)

        self.observations = observations

    def run_location(self, observations=None, calculate_rays=True,
                     delete_output_files=True, event=None,
                     evaluation_mode: str = 'automatic',
                     evaluation_status: str = 'preliminary',
                     multithreading=False):

        import subprocess

        if event is not None:
            observations = Observations.from_event(event=event)

        if (observations is None) and (self.observations is None):
            raise ValueError('The current run does not contain travel time'
                             'observations. Observations should be added to '
                             'the current run using the add_observations '
                             'method.')

        elif observations is not None:
            self.__add_observations__(observations)

        self.write_control_file()

        cmd = ['NLLoc', self.control_file]
        process = subprocess.Popen(cmd, stdout=subprocess.PIPE)

        logger.info('locating event using NonLinLoc')
        t0 = time()
        output, error = process.communicate()
        t1 = time()
        logger.info(f'done locating event in {t1 - t0:0.2f} seconds')

        if event is not None:
            event_time = event.time
        else:
            p_times = []
            for pick in observations.picks:
                p_times.append(pick.time)
            event_time = np.min(p_times)

        if not (self.output_file_path / 'last.hyp').exists():
            logger.error(f'event location failed for event {event_time}!')
        t, x, y, z = read_hypocenter_file(self.output_file_path / 'last.hyp')

        scatters = read_scatter_file(self.output_file_path / 'last.scat')

        rays = None
        if calculate_rays:
            logger.info('calculating rays')
            t0_ray = time()
            rays = self.travel_times.ray_tracer(np.array([x, y, z]),
                                                multithreading=multithreading)
            t1_ray = time()
            logger.info(f'done calculating rays in {t1_ray - t0_ray:0.2f} '
                        f'seconds')

        if delete_output_files:
            for fle in self.observation_path.glob('*'):
                fle.unlink()
            self.observation_path.rmdir()

            output_dir = self.output_file_path

            for fle in output_dir.glob('*'):
                fle.unlink()
            output_dir.rmdir()

            for fle in self.output_file_path.parent.glob('*'):
                fle.unlink()

            self.output_file_path.parent.rmdir()

        result = NLLOCResult(np.array([x, y, z]), t, scatters, rays,
                             observations, evaluation_mode, evaluation_status)

        return result

    def rays(self, hypocenter_location):
        return self.travel_times.ray_tracer(hypocenter_location)

    @property
    def nlloc_files(self):
        return NllocInputFiles(self.observation_file,
                               self.travel_time_grid_location /
                               self.network_code,
                               self.output_file_path / self.network_code)

    @property
    def control(self):

        if self.srces is None:
            raise ValueError('The project does not contain sensors or '
                             'inventory. Sensors (srces) or inventory '
                             'information can be added using the add_srces or'
                             'add_inventory methods.')

        if self.observations is None:
            raise ValueError('The current run does not contain travel time'
                             'observations. Observations should be added to '
                             'the current run using the add_observations '
                             'method.')

        observations = str(self.observations)

        ctrl = ''
        ctrl += str(self.control_template['control']) + '\n'
        ctrl += str(self.control_template['transformation']) + '\n\n'

        if self.control_template['locsig'] is not None:
            ctrl += self.control_template['locsig']

        if self.control_template['loccom'] is not None:
            ctrl += self.control_template['loccom'] + '\n'

        ctrl += str(self.srces) + '\n'

        ctrl += str(self.nlloc_files) + '\n'

        ctrl += str(self.control_template['locsearch'])
        ctrl += str(self.control_template['locmeth'])
        ctrl += str(self.control_template['locgau']) + '\n'

        ctrl += str(self.control_template['locqual2err'])

        if self.p_velocity is not None:
            ctrl += str(LocGrid.init_from_grid(self.p_velocity))
        else:
            raise ValueError('Cannot initialize the LocGrid, the velocity '
                             'grids are not defined')

        return ctrl



from pathlib import Path
import pickle
from uquake.core.inventory import read_inventory
from uquake.core.logging import logger
from uquake.nlloc import (Srces)
from uquake.core.event import ConfidenceEllipsoid, OriginUncertainty
from uquake.grid import nlloc as nlloc_grid
from time import time
import numpy as np
import os
import shutil
from ..settings.settings import Settings


class ProjectManager(object):

    inventory_file_name = 'inventory.xml'

    def __init__(self, path, project_name, network_code, use_srces=False):
        """
        Interface to manage project providing an interface to selected
        components of the NonLinLoc software by Anthony Lomax.

        :param path: base path
        :type path: str
        :param project_name: project name or id
        :type project_name: str
        :param network_code: network name or id
        :type network_code: str
        :program use_srces: if True use the srces files instead of the the
        inventory file should both files be present (default=False)
        :Example:
        >>> from uquake.grid import nlloc as nlloc_grid
        >>> from uquake.nlloc import nlloc
        # initialize the project with the root path of the project,
        the project and the network names
        >>> project_name = 'test'
        >>> network_code = 'test'
        >>> root_path=''
        >>> pm = nlloc.ProjectManager(root_path, project_name, network_code)
        # this will initialize the project and create a specific directory
        # structure if the directories do not exist.

        # add an inventory file to the project
        >>> path_to_uquake_inventory_file = 'PATH/TO/THE/INVENTORY/FILE.xml'

        ..note :: the uquake inventory object inherit from the obspy inventory
                  and
                  behaves in a very similar way. It, however, differs from
                  its parent has it implements properties specific to local
                  earthquake who are either expressed using UTM coordinates
                  system or a local coordinates system that cannot necessarily
                  be translated to latitude and longitude. Beyond this minor
                  difference, uquake inventory structure also differs from
                  the standard inventory and implements a slightly different
                  hierarchy more representative of mine monitoring systems.
                  The uquake Inventory object hierarchy is as follows:
                 1. Inventory
                    1.1. Networks: A useismic network. a network includes at
                                   least one data acquisition station.
                        1.1.1 Stations: A place where the data acquisition is
                                        performed. For instance, station
                                        usually includes power, communication
                                        and data acquisition equipment. One
                                        or more sensor can be connected to a
                                        station.
                            1.1.1.1 Sensors: A instrument converting a
                                             physical phenomenon to data either
                                             digital or analog. A sensor
                                             comprises one or more channel.
                                1.1.1.1.1 Channels: A channel is a the smallest
                                                    unit of measuring.


        >>> inventory = read_inventory(path_to_uquake_inventory_file)
        >>> pm.add_inventory(inventory)

        # alternatively, sensors can be added to the using the srces object.
        # Sensors can be added this way from a nlloc.nlloc.Srces object using
        the .add_srces method.
        ..note:: srces stands for sources and it is the nomenclature used in
                 NonLinLoc. This might be a soure of confusion for the users.
                 In addition, what NonLinLoc refers to as station is called a
                 sensors in this context.

        # srces object can be constructed from an inventory as follows:
        >>> srces = Srces.from_inventory(inventory)
        # alternatively, each sensors can be added individually using the
        .add_sensor method. As follows:
        >>> srces = Srces()
        >>> x = 250
        >>> y = 250
        >>> z = 250
        >>> elevation = 0
        >>> srces.add_sensor('sensor label', x, y, z, elev=elevation)

        # Srces can be added to the project as follows:
        >>> pm.add_srces(srces)

        # add the velocity models to the project. P- and S-wave velocity models
        can be added separately from a nlloc.grid.VelocityGrid3D
        using the .add_velocity method or from a nlloc.grid.VelocityEnsemble
        object using the .add_velocities method

        >>> origin = [0, 0, 0]
        >>> spacing = [25, 25, 25]
        >>> dimensions = [100, 100, 100]

        >>> nlloc_grid.VelocityGrid3D()
        >>> vp = 5000  # P-wave velocity in m/s
        >>> vs = 3500  # S-wave velocity in m/s
        >>> p_velocity = nlloc_grid.VelocityGrid3D(network_code, dimensions,
        >>>                                        origin, spacing, phase='P',
        >>>                                        value=5000)
        >>> s_velocity = nlloc_grid.VelocityGrid3D(network_code, dimensions,
        >>>                                        origin, spacing, phase='S',
        >>>                                        value=5000 )
        >>> pm.add_velocity(p_velocity)
        >>> pm.add_velocity(s_velocity)
        # Alternatively
        >>> velocities = nlloc_grid.VelocityGridEnsemble(p_velocity,
        >>>                                              s_velocity)
        >>> pm.add_velocities(velocities)
        # Adding a velocity model of the velocity models triggers the
        # calculation of the travel time grids.
        # It is possible to manually trigger the calculation of the travel
        # time grid by invoking
        >>> pm.init_travel_time_grid()
        # this should not, however, be required.

        # prior to running the location, NonLinLoc need to be configured.
        # configuring NonLinLoc can be done using the nlloc.nlloc.NonLinLoc
        # object
        >>> nonlinloc = nlloc.NonLinLoc()
        # this will initialize the nonlinloc object sith default value. Those
        # value have been used to locate useismic events in a volumes of
        # approximately 3000 m x 3000 m x 1500 m. The parameters should be
        # provide adequate results for volumes of similar scale but would need
        # to be adapted to smaller or larger volumes.

        """

        self.project_name = project_name
        self.network_code = network_code
        self.root_directory = Path(path) / project_name / network_code
        # create the directory if it does not exist
        self.root_directory.mkdir(parents=True, exist_ok=True)

        self.inventory_location = self.root_directory / 'inventory'
        self.inventory_location.mkdir(parents=True, exist_ok=True)
        self.inventory_file = self.inventory_location / 'inventory.xml'
        self.srces_file = self.inventory_location / 'srces.pickle'

        # SETTINGS

        self.config_location = self.root_directory / 'config'

        self.config_location.mkdir(parents=True, exist_ok=True)

        self.settings_file = self.config_location / 'settings.toml'
        self.service_setting_file = self.config_location / \
                                    'services_settings.toml'

        if not self.settings_file.is_file():
            settings_template = Path(os.path.realpath(__file__)).parent / \
                                    '../settings/settings_template.toml'

            shutil.copyfile(settings_template, self.settings_file)

        if not self.service_setting_file.is_file():
            settings_path = Path(os.path.realpath(__file__)).parent.parent
            settings_template = settings_path / \
                                'settings/services_settings_template.toml'

            shutil.copyfile(settings_template, self.config_location /
                            'service_settings.toml')

        self.settings = Settings(str(self.config_location))

        self.srces = None
        self.inventory = None

        if (not self.inventory_file.exists()) and \
                (not self.srces_file.exists()):
            logger.warning('the project does not contain an inventory file nor'
                           'an srces file. to add an inventory file use '
                           'the add_inventory method. Alternatively, sensors '
                           'information can be provided using an Srces object.'
                           'Note, however, that an Srces object only contains'
                           'the sensors location information. When '
                           'the inventory file is present, the Srces object '
                           'is automatically constructed from the inventory.'
                           'A Srces object can be added using the add_srces '
                           'method. If both the Srces and inventory files are'
                           'present in the project directory, the inventory'
                           'file takes precedence and the srces file is'
                           'ignored unless')

        elif self.srces_file.exists() and use_srces:
            with open(self.srces_file, 'rb') as srces_file:
                logger.info('srces will be read from the file and not build'
                            'from the inventory file')
                self.srces = pickle.load(srces_file)

            if self.inventory_file.exists():
                self.inventory = read_inventory(str(self.inventory_file))

        elif self.inventory_file.exists():
            self.inventory = read_inventory(str(self.inventory_file))
            self.srces = Srces.from_inventory(self.inventory)
            logger.info('srces will be build from the inventory file. The '
                        'srces.pickle file will be replaced.')
            with open(self.srces_file, 'wb') as srces_file:
                pickle.dump(self.srces, srces_file)

        elif self.srces_file.exists():
            logger.info('no inventory file in the project. srces file will be '
                        'used instead.')
            with open(self.srces_file, 'rb') as srces_file:
                self.srces = pickle.load(srces_file)

        self.velocity_grid_location = self.root_directory / 'velocities'
        self.velocity_grid_location.mkdir(parents=True, exist_ok=True)

        p_vel_base_name = nlloc_grid.VelocityGrid3D.get_base_name(network_code, 'P')
        self.p_velocity_file = self.velocity_grid_location / p_vel_base_name
        p_files = self.velocity_grid_location.glob(f'{p_vel_base_name}*')
        if len(list(p_files)) == 0:
            logger.warning('the project does not contain a p-wave velocity '
                           'model. to add a p-wave velocity model to the '
                           'project please use the add_velocity or '
                           'add_velocities methods.')
            self.p_velocity = None

        else:
            self.p_velocity = nlloc_grid.read_grid(p_vel_base_name,
                                                   path=str(
                                                 self.velocity_grid_location))

        s_vel_base_name = nlloc_grid.VelocityGrid3D.get_base_name(self.network_code,
                                                            'S')
        self.s_velocity_file = self.velocity_grid_location / s_vel_base_name
        s_files = self.velocity_grid_location.glob(f'{s_vel_base_name}*')
        if len(list(s_files)) == 0:
            logger.warning('the project does not contain a s-wave velocity '
                           'model. to add a s-wave velocity model to the '
                           'project please use the add_velocity or '
                           'add_velocities methods.')
            self.s_velocity = None
        else:
            self.s_velocity = nlloc_grid.read_grid(s_vel_base_name,
                                                   path=str(
                                                 self.velocity_grid_location))

        self.velocities = None
        if (self.p_velocity is not None) and (self.s_velocity is not None):
            self.velocities = nlloc_grid.VelocityGridEnsemble(self.p_velocity,
                                                              self.s_velocity)

        self.travel_time_grid_location = self.root_directory / 'times'
        self.travel_time_grid_location.mkdir(parents=True, exist_ok=True)
        file_list = list(self.travel_time_grid_location.glob('*time*'))
        if len(file_list) == 0:
            logger.warning('the project does not contain travel time grids. '
                           'to initialize the travel-time grid use the '
                           'init_travel_time_grid method. Note that '
                           'this require the project to contain both '
                           'an inventory and a velocities files.')

            self.travel_times = None

        else:
            self.travel_times = nlloc_grid.TravelTimeEnsemble.from_files(
                self.travel_time_grid_location)

    def init_travel_time_grid(self):
        """
        initialize the travel time grids
        """
        logger.info('initializing the travel time grids')
        t0 = time()
        seeds = self.srces.locs
        seed_labels = self.srces.labels

        if self.srces is None:
            return

        if self.p_velocity:
            tt_gs_p = self.p_velocity.to_time_multi_threaded(seeds,
                                                             seed_labels)

        if self.s_velocity:
            tt_gs_s = self.s_velocity.to_time_multi_threaded(seeds,
                                                             seed_labels)

        if self.p_velocity and self.s_velocity:
            self.travel_times = tt_gs_p + tt_gs_s

        elif self.p_velocity:
            logger.warning('s-wave velocity model is not set, travel-times '
                           'will'
                           'will be generated for ')
            self.travel_times = tt_gs_p

        elif self.s_velocity:
            self.travel_times = tt_gs_s

        else:
            return

        # cleaning the directory before writing the new files

        for fle in self.travel_time_grid_location.glob('*time*'):
            fle.unlink(missing_ok=True)

        self.travel_times.write(self.travel_time_grid_location)
        t1 = time()
        logger.info(f'done initializing the travel time grids in '
                    f'{t1 - t0:0.2f} seconds')

    def add_inventory(self, inventory, create_srces_file: bool = True,
                      initialize_travel_time: bool = True):
        """
        adding a inventory object to the project
        :param inventory: station xml inventory object
        :type inventory: uquake.core.inventory.Inventory
        :param create_srces_file: if True create or replace the srces file
        (Default: True)
        :param initialize_travel_time: if True the travel time grids are
        initialized (default: True)
        :return:
        """

        inventory.write(str(self.inventory_file))
        self.srces = Srces.from_inventory(inventory)
        if create_srces_file:
            with open(self.srces_file, 'wb') as srces_file:
                pickle.dump(self.srces, srces_file)

        if initialize_travel_time:
            self.init_travel_time_grid()
        else:
            logger.warning('the travel time grids will not be initialized, '
                           'the inventory and the travel time grids might '
                           'be out of sync. To initialize the travel time '
                           'grids make sure initialize_travel_time is set to '
                           'True')

    def add_srces(self, srces, force=False, initialize_travel_time=True):
        """
        add a list of sources to the projects
        :param srces: list of sources or sensors
        :param force: force the insertion of the srces object if an inventory
        file is present
        :param initialize_travel_time: if True, initialize the travel time
        grid
        :type srces: Srces

        ..warning:: travel time should be initialized when the sensors/srces
        are updated. Not doing so, may cause the sensors/source and the
        travel time grids to be incompatible.
        """

        if not isinstance(srces, Srces):
            raise TypeError(f'Expecting type {Srces}, given '
                            f'{type(srces)}')

        if self.inventory is not None:
            logger.warning('The project already has an inventory file!')
            if not force:
                logger.warning('exiting...')
                return
            else:
                logger.warning('the force flag value is True, srces object '
                               'will be added.')

        self.srces = srces
        with open(self.srces_file, 'wb') as srces_file:
            pickle.dump(self.srces, srces_file)

        if initialize_travel_time:
            self.init_travel_time_grid()
        else:
            logger.warning('the travel time grids will not be initialized, '
                           'the inventory and the travel time grids might '
                           'be out of sync. To initialize the travel time '
                           'grids make sure initialize_travel_time is set to '
                           'True')

        self.init_travel_time_grid()

    def add_velocities(self, velocities, initialize_travel_time=True):
        """
        add P- and S-wave velocity models to the project
        :param velocities: velocity models
        :type velocities: nlloc.grid.VelocityEnsemble
        :param initialize_travel_time: if True, initialize the travel time
        grid

        ..warning:: travel time should be initialized when the sensors/srces
        are updated. Not doing so, may cause the sensors/source and the
        travel time grids to be incompatible.
        """

        # velocities.write(path=self.velocity_grid_location)

        self.velocities = velocities

        for key in velocities.keys():
            self.add_velocity(velocities[key],
                              initialize_travel_time=False)

        if initialize_travel_time:
            self.init_travel_time_grid()
        else:
            logger.warning('the travel time grids will not be initialized, '
                           'the inventory and the travel time grids might '
                           'be out of sync. To initialize the travel time '
                           'grids make sure initialize_travel_time is set to '
                           'True')

    def add_velocity(self, velocity, initialize_travel_time=True):
        """
        add P- or S-wave velocity model to the project
        :param velocity: p-wave velocity model
        :type velocity: nlloc.grid.VelocityGrid3D
        :param initialize_travel_time: if true initialize the travel time grids

        ..warning:: travel time should be initialized when the sensors/srces
        are updated. Not doing so, may cause the sensors/source and the
        travel time grids to be incompatible.
        """

        velocity.write(path=self.velocity_grid_location)

        if velocity.phase.upper() == 'P':
            self.p_velocity = velocity

        else:
            self.s_velocity = velocity

        if initialize_travel_time:
            self.init_travel_time_grid()
        else:
            logger.warning('the inventory and the travel time grids might'
                           'be out of sync.')

    def clean_project(self):
        pass




import json
import urllib.parse
from io import BytesIO

import requests
from dateutil import parser
from obspy import UTCDateTime
from libraries.uquake.uquake.core.event import Catalog
from obspy.core.util.attribdict import AttribDict

from uquake.core.logging import logger
from uquake.core import read
from uquake.core.event import Ray, read_events
from uuid import uuid4
from uquake.core.decorators import deprecated

timeout = 200


class RequestRay(AttribDict):
    def __init__(self, json_data):
        super(RequestRay, self).__init__(json_data)
        self.ray = Ray(self.nodes)
        del self.nodes


class RequestEvent:
    def __init__(self, ev_dict):
        for key in ev_dict.keys():
            if 'time' in key:
                if key == 'timezone':
                    continue

                if type(ev_dict[key]) is not str:
                    setattr(self, key, ev_dict[key])

                    continue
                setattr(self, key, UTCDateTime(parser.parse(ev_dict[key])))
            else:
                setattr(self, key, ev_dict[key])

    def get_event(self):
        event_file = requests.request('GET', self.event_file, timeout=timeout)

        return read_events(event_file.content, format='QUAKEML')

    def get_waveforms(self):
        if self.waveform_file is None:
            logger.warning(f'No waveform object associated with event '
                           f'{self.event_resource_id}')
            return None
        waveform_file = requests.request('GET', self.waveform_file, 
                                         timeout=timeout)
        byte_stream = BytesIO(waveform_file.content)

        return read(byte_stream, format='MSEED')

    def get_context_waveforms(self):
        if self.waveform_context_file is None:
            logger.warning(f'No context trace associated with event '
                           f'{self.event_resource_id}')
            return None
        waveform_context_file = requests.request('GET',
                                                 self.waveform_context_file,
                                                 timeout=timeout)
        byte_stream = BytesIO(waveform_context_file.content)

        return read(byte_stream)

    def get_variable_length_waveforms(self):
        if self.variable_size_waveform_file is None:
            logger.warning(f'No variable length waveform associated with '
                           f'event {self.event_resource_id}')
            return None

        variable_length_waveform_file = requests.request('GET',
                                                         self.variable_size_waveform_file,
                                                         timeout=timeout)

        byte_stream = BytesIO(variable_length_waveform_file.content)

        return read(byte_stream)

    def select(self):
        # select by different attributes
        # TODO write this function :-)
        pass

    def keys(self):
        return self.__dict__.keys()

    def __repr__(self):
        outstr = "\n"

        for key in self.keys():
            outstr += '%s: %s\n' % (key, self.__dict__[key])
        outstr += "\n"

        return outstr


# class SeismicClient:
#     def __init__(self, api_base_url, username, password,
#                  microquake_event_type=True):
#         configuration = seismic_client.Configuration()
#         configuration.username = username
#         configuration.password = password
#         self.api_base_url = api_base_url
#         self.username = username
#         self.password = password
#         self.api_instance = seismic_client.ApiApi(
#              seismic_client.ApiClient(configuration))
#
#     @staticmethod
#     def event_detail():
#         return seismic_client.EventDetail()
#
#     def events_list(self, start_time=None,
#                     end_time=None,
#                     microquake_event_type=True,
#                     **kwargs):
#         """
#         get the list of event from start_time to end_time
#         :param start_time: start_time in UTC expressed as
#         "YYYY-MM-DDTHH:MM:SS.MS" or a UTCDateTime, or a datetime
#         explicitely providing this paramter will override the time_utc_after
#         :param end_time: end_time in UTC expressed as "YYYY-MM-DDTHH:MM:SS.MS"
#         or a UTCDateTime, or a datetime
#         explicitely providing this paramter will override the time_utc_before
#         :param microquake_event_type: if true use microquake event types else
#         use Quakeml event types
#         :param kwargs:
#         :return:
#         """
#
#         if 'event_type' in kwargs.keys():
#             if microquake_event_type:
#                 event_types_lut = get_event_types(self.api_base_url,
#                                                   username=self.username,
#                                                   password=self.password)
#                 kwargs['event_type'] = event_types_lut[kwargs['event_type']]
#
#         if start_time:
#             kwargs['time_utc_after'] = str(start_time)
#         if end_time:
#             kwargs['time_utc_before'] = str(end_time)
#
#         events = []
#
#         try:
#             response = self.api_instance.api_v1_events_list(**kwargs)
#         except Exception as e:
#             logger.error(e)
#             return {}, events
#
#         for event in response.results:
#             events.append(RequestEvent(event.to_dict()))
#         query = response.next
#
#         while query:
#             re = requests.get(query, timeout=timeout)
#             if not re:
#                 break
#             response = re.json()
#             logger.info(f"page {response['current_page']} of "
#                         f"{response['total_pages']}")
#
#             query = response['next']
#
#             for event in response['results']:
#                 events.append(RequestEvent(event))
#
#         return response, events
#
#     def events_read(self, event_resource_id):
#         encoded_id = encode(event_resource_id)
#         if self.api_base_url[-1] != '/':
#             self.api_base_url + '/'
#         encoded_id = event_resource_id
#         url = self.api_base_url + 'events/' + encoded_id
#
#         try:
#             # api_response = self.api_instance.api_v1_events_read(encoded_id)
#             api_response = requests.get(url,
#                                         auth=(self.username, self.password))
#         except Exception as e:
#             logger.error(e)
#             return None, None
#
#         if not api_response:
#             return api_response, None
#
#         event = None
#         if api_response.json():
#             event = RequestEvent(api_response.json())
#
#         return api_response, event
#
#     def events_partial_update(self, event_resource_id, **kwargs):
#
#         data = kwargs
#         encoded_id = urllib.parse.quote(event_resource_id, safe='')
#         url = self.api_base_url + 'events/' + encoded_id
#         resp = requests.patch(url, json=data, auth=(self.username,
#                                                     self.password))
#         logger.info(resp)
#         return(resp)
#
#     def post_data_from_objects(self, network, event_id=None, cat=None,
#                                 stream=None, context=None, variable_length=None,
#                                 tolerance=0.5, send_to_bus=False):
#         response = post_data_from_objects(self.api_base_url, network,
#                                           event_id=event_id,
#                                           cat=cat, stream=stream,
#                                           context=context,
#                                           variable_length=variable_length,
#                                           tolerance=tolerance,
#                                           send_to_bus=send_to_bus,
#                                           username=self.username,
#                                           password=self.password)
#         return response
#
#         # try:
#         #     response = self.api_instance.api_v1_events_partial_update_0(body,
#         #                                  event_resource_id)
#         # except Exception as e:
#         #     logger.error(e)
#         # return response


def encode(resource_id):
    return urllib.parse.quote(resource_id, safe='')


def post_data_from_files(api_base_url, network, event_id=None, event_file=None,
                         mseed_file=None, mseed_context_file=None,
                         variable_length_stream_file=None,
                         tolerance=0.5):
    """
    Build request directly from objects
    :param api_base_url: base url of the API
    :param network: network code
    :param event_id: event_id (not required if an event is provided)
    :param event_file: path to a QuakeML file
    :param stream_file: path to a mseed seismogram file
    :param context_stream_file: path to a context seismogram file
    :param tolerance: Minimum time between an event already in the database
    and this event for this event to be inserted into the database. This is to
    avoid duplicates. event with a different
    <event_id> within the <tolerance> seconds of the current object will
    not be inserted. To disable this check set <tolerance> to None.
    :return: same as build_request_data_from_bytes
    """

    event = None,
    stream = None
    context_stream = None

    # read event

    if event_file is not None:
        event = read_events(event_file)

    # read waveform

    if mseed_file is not None:
        stream = read(mseed_file, format='MSEED')

    # read context waveform

    if mseed_context_file is not None:
        context_stream = read(mseed_context_file, format='MSEED')

    # read variable length waveform

    if variable_length_stream_file is not None:
        variable_length_stream = read(variable_length_stream_file,
                                      format='MSEED')

    return post_data_from_objects(api_base_url, network, event_id=event_id,
                                  cat=event, stream=stream,
                                  context=context_stream,
                                  variable_length=variable_length_stream,
                                  tolerance=tolerance)


def post_data_from_objects(api_base_url, network, event_id=None, cat=None,
                           stream=None, context=None, variable_length=None,
                           tolerance=0.5, send_to_bus=False, username=None,
                           password=None):
    """
    Build request directly from objects
    :param api_base_url: base url of the API
    :param event_id: event_id (not required if an event is provided)
    :param cat: microquake.core.event.Event or a
    microquake.core.event.Catalog containing a single event. If the catalog
    contains more than one event, only the first event, <catalog[0]> will be
    considered. Use catalog with caution.
    :param stream: event seismogram (microquake.core.Stream.stream)
    :param context: context seismogram trace (
    microquake.core.Stream.stream)
    :param variable_length: variable length seismogram trace (
    microquake.core.Stream.stream)
    :param tolerance: Minimum time between an event already in the database
    and this event for this event to be inserted into the database. This is to
    avoid duplicates. event with a different
    <event_id> within the <tolerance> seconds of the current object will
    not be inserted. To disable this check set <tolerance> to None.
    :param logger: a logging.Logger object
    :return: same as build_request_data_from_bytes
    """

    api_url = api_base_url + "events"

    if type(cat) is Catalog:
        cat = cat[0]
        logger.warning('a <microquake.core.event.Catalog> object was '
                       'provided, only the first element of the catalog will '
                       'be used, this may lead to an unwanted behavior')

    if cat is not None:
        event_time = cat.preferred_origin().time
        event_resource_id = str(cat.resource_id)
    else:
        if event_id is None:
            logger.warning('A valid event_id must be provided when no '
                           '<event> is not provided.')
            logger.info('exiting')

            return

        re = get_event_by_id(api_base_url, event_id)

        if re is None:
            logger.warning('request did not return any event with the '
                           'specified event_id. A valid event_id or an event '
                           'object must be provided to insert a stream or a '
                           'context stream into the database.')
            logger.info('exiting')

            return

        event_time = re.time_utc
        event_resource_id = str(re.event_resource_id)

    # data['event_resource_id'] = event_resource_id

    logger.info('processing event with resource_id: %s' % event_resource_id)

    base_event_file_name = str(event_time)
    files = {}
    # Test if there is an event within <tolerance> seconds in the database
    # with a different <event_id>. If so abort the insert

    if tolerance is not None:
        start_time = event_time - tolerance
        end_time = event_time + tolerance
        re_list = get_events_catalog(api_base_url, start_time, end_time)

        if re_list:
            logger.warning('Event found within % seconds of current event'
                           % tolerance)
            logger.warning('The current event will not be inserted into the'
                           ' data base')

            return

    files = prepare_data(cat=cat, stream=stream, context=context,
                         variable_length=variable_length)

    return post_event_data(api_url, network, event_resource_id, files,
                           send_to_bus=send_to_bus, username=username,
                           password=password)


def prepare_data(cat=None, stream=None, context=None, variable_length=None):

    files = {}
    base_event_file_name = str(uuid4())

    if cat is not None:
        logger.info('preparing event data')
        event_bytes = BytesIO()
        cat.write(event_bytes, format='QUAKEML')
        event_file_name = base_event_file_name + '.xml'
        event_bytes.name = event_file_name
        event_bytes.seek(0)
        files['event_file'] = event_bytes
        logger.info('done preparing event data')

    if stream is not None:
        logger.info('preparing waveform data')
        mseed_bytes = BytesIO()
        stream.write(mseed_bytes, format='MSEED')
        mseed_file_name = base_event_file_name + '.mseed'
        mseed_bytes.name = mseed_file_name
        mseed_bytes.seek(0)
        files['waveform_file'] = mseed_bytes
        logger.info('done preparing waveform data')

    if context is not None:
        logger.info('preparing context waveform data')
        mseed_context_bytes = BytesIO()
        context.write(mseed_context_bytes, format='MSEED')
        mseed_context_file_name = base_event_file_name + '.context_mseed'
        mseed_context_bytes.name = mseed_context_file_name
        mseed_context_bytes.seek(0)
        files['waveform_context_file'] = mseed_context_bytes
        logger.info('done preparing context waveform data')

    if variable_length is not None:
        logger.info('preparing variable length waveform data')
        mseed_variable_bytes = BytesIO()
        variable_length.write(mseed_variable_bytes, format='MSEED')
        mseed_variable_file_name = base_event_file_name + '.variable_mseed'
        mseed_variable_bytes.name = mseed_variable_file_name
        mseed_variable_bytes.seek(0)
        files['variable_size_waveform_file'] = mseed_variable_bytes
        logger.info('done preparing variable length waveform data')

    return files


def get_event_types(api_base_url, username=None, password=None):

    if api_base_url[-1] != '/':
        api_base_url += '/'

    url = api_base_url + 'inventory/microquake_event_types'
    if username:
        response = requests.get(url, auth=(username, password))
    else:
        response = requests.get(url, timeout=timeout)

    if not response:
        raise ConnectionError('API Connection Error')

    data = json.loads(response.content)
    dict_out = {}
    for d in data:
        dict_out[d['microquake_type']] = d['quakeml_type']

    return dict_out


def post_event_data(api_base_url, network, event_resource_id, request_files,
                    send_to_bus=False, username=None, password=None):
    # removing id from URL as no longer used
    # url = api_base_url + "%s/" % event_resource_id
    url = api_base_url
    logger.info('posting data on %s' % url)

    event_resource_id = encode(event_resource_id)

    if username is None:
        result = requests.post(url, data={"send_to_bus": send_to_bus},
                               files=request_files, timeout=timeout)
    else:
        result = requests.post(url, data={"send_to_bus": send_to_bus},
                               files=request_files, timeout=timeout,
                               auth=(username, password))
    logger.info(result)

    return result


def put_data_from_objects(api_base_url, network, cat=None, stream=None,
                          context=None, variable_length=None, username=None,
                          password=None):

    event_id = cat[0].resource_id.id

    base_url = api_base_url
    if base_url[-1] == '/':
        base_url = base_url[:-1]

    # check if the event type and event status has changed on the API

    # re = get_event_by_id(api_base_url, event_id)
    # try:
    #     cat[0].event_type = re.event_type
    #     cat[0].preferred_origin().evaluation_status = re.status
    # except AttributeError as e:
    #     logger.error(e)

    url = f'{base_url}/events/{event_id}'

    files = prepare_data(cat=cat, stream=stream, context=context,
                         variable_length=variable_length)

    logger.info(f'attempting to PUT catalog for event {event_id}')

    if username is None:
        response = requests.patch(url, files=files, timeout=timeout)
    else:
        response = requests.patch(url, files=files, timeout=timeout,
                                  auth=(username, password))

    logger.info(f'API responded with {response.status_code} code')
    return response


@deprecated
def get_events_catalog(api_base_url, start_time, end_time,
                       status='accepted', event_type=''):
    """
    return a list of events
    :param api_base_url:
    :param start_time:
    :param end_time:
    :param event_type:  example seismic_event,blast,drilling noise,open pit
    blast,quarry blast... etc
    :param status: Event status, accepted, rejected, accepted,rejected
    :return:
    """
    url = api_base_url + "catalog"

    # request work in UTC, time will need to be converted from whatever
    # timezone to UTC before the request is built.

    querystring = {"start_time": start_time, "end_time": end_time, "status":
                   status, "type": event_type}

    response = requests.request("GET", url, params=querystring,
                                timeout=timeout).json()

    events = []

    for event in response:
        events.append(RequestEvent(event))

    return events


def get_event_by_id(api_base_url, event_resource_id):
    # smi:local/e7021615-e7f0-40d0-ad39-8ff8dc0edb73
    url = api_base_url + "events/"
    # querystring = {"event_resource_id": event_resource_id}

    event_resource_id = encode(event_resource_id)
    response = requests.request("GET", url + event_resource_id,
                                timeout=timeout)

    if response.status_code != 200:
        return None

    return RequestEvent(json.loads(response.content))


def post_signal_quality(
        api_base_url, request_data, username=None, password=None
):
    session = requests.Session()
    session.trust_env = False

    if username is None:
        return session.post("{}signal_quality".format(api_base_url),
                            json=request_data)
    else:
        return session.post("{}signal_quality".format(api_base_url),
                            json=request_data, auth=(username, password))


def reject_event(api_base_url, event_id):
    session = requests.Session()
    session.trust_env = False

    event_id = encode(event_id)

    return session.post("{}{}/interactive/reject".format(api_base_url,
                                                         event_id))


def get_catalog(api_base_url, start_time, end_time, event_type=None,
                status=None, **kwargs):
    """
    get the event catalogue from the API
    :param api_base_url: API base url
    :param start_time: start time as datetime if not time aware, UTC is
    assumed
    :param end_time: end time as datetime if not time aware, UTC is assumed
    :param time_zone:
    :param event_type: microquake event type
    :param status: event status acceptable values are ('preliminary',
    'reviewed', 'confirmed', 'final', 'rejected', accepted'). 'accepted'
    encompasses all status with the exception of 'rejected'
    :return: a RequestEvent object
    """

    if api_base_url[-1] != '/':
        api_base_url += '/'

    api_base_url += 'events'

    event_types = get_event_types(api_base_url)
    obs_event_type = None

    request_dict = {'time_utc_after': str(start_time),
                    'time_utc_before': str(end_time)}

    if event_type is not None:
        try:
            request_dict['event_type'] = event_types[event_type]
        except KeyError:
            logger.error(f'event type: {event_type} does not appear to be a '
                         f'valid event type for your system')

            raise KeyError
    else:
        event_type_str = ''
        for key in event_types.keys():
            event_type_str += f'{event_types[key]},'
        event_type_str = event_type_str[:-1]
        request_dict['event_type'] = event_type_str

    if status is not None:
        request_dict['status'] = status

    events = []

    tmp = urllib.parse.urlencode(request_dict)
    query = f'{api_base_url}?{tmp}'

    while query:
        re = requests.get(query, timeout=timeout)
        # from ipdb import set_trace; set_trace()()
        if not re:
            # logger.info('The API catalogue does not contain any events that'
            #             'corresponds to the request')
            break
        response = re.json()
        logger.info(f"page {response['current_page']} of "
                    f"{response['total_pages']}")

        query = response['next']

        for event in response.results:
            events.append(RequestEvent(event))

    return events


def get_catalog_modified(api_base_url, modification_start_time,
                         modification_end_time, event_type=None,
                         status=None):
    """
    get the event catalogue from the API
    :param api_base_url: API base url
    :param modification_start_time: start time as datetime if not time aware, UTC is
    assumed
    :param modification_end_time: end time as datetime if not time aware, UTC is assumed
    :param event_type: microquake event type
    :param status: event status acceptable values are ('preliminary',
    'reviewed', 'confirmed', 'final', 'rejected', accepted'). 'accepted'
    encompasses all status with the exception of 'rejected'
    :return: a RequestEvent object
    """

    if api_base_url[-1] != '/':
        api_base_url += '/'

    api_base_url += 'events'

    event_types = get_event_types(api_base_url)
    obs_event_type = None

    request_dict = {'modification_timestamp_after': str(
                     modification_start_time),
                    'modification_timestamp_before': str(
                     modification_end_time)}

    if event_type is not None:
        try:
            request_dict['event_type'] = event_types[event_type]
        except KeyError:
            logger.error(f'event type: {event_type} does not appear to be a '
                         f'valid event type for your system')

            raise KeyError
    else:
        event_type_str = ''
        for key in event_types.keys():
            event_type_str += f'{event_types[key]},'
        event_type_str = event_type_str[:-1]
        request_dict['event_type'] = event_type_str

    if status is not None:
        request_dict['status'] = status

    events = []

    tmp = urllib.parse.urlencode(request_dict)
    query = f'{api_base_url}?{tmp}'

    while query:
        re = requests.get(query, timeout=timeout)
        # from ipdb import set_trace; set_trace()()
        if not re:
            # logger.info('The API catalogue does not contain any events that'
            #             'corresponds to the request')
            break
        response = re.json()
        logger.info(f"page {response['current_page']} of "
                    f"{response['total_pages']}")

        query = response['next']

        for event in response['results']:
            events.append(RequestEvent(event))

    return events


def get_magnitude(api_base_url, magnitude_id):

    if api_base_url[-1] != '/':
        api_base_url += '/'

    request_url = f'{api_base_url}magnitudes/{magnitude_id}'

    # requests_dict = {'event_id': event_id,
    #                  'origin_id': origin_id}

    # tmp = urllib.parse.urlencode(requests_dict)
    # query = f'{_url}'

    return requests.get(request_url)



import os
from dynaconf import LazySettings
from pathlib import Path


class Settings(LazySettings):
    def __init__(self, settings_location, settings_file='settings.toml'):
        """
        Init function currently just initializes the object allowing
        """

        config_dir = Path(settings_location)

        dconf = {}
        dconf.setdefault('ENVVAR_PREFIX_FOR_DYNACONF', 'UQUAKE')

        env_prefix = '{0}_ENV'.format(
            dconf['ENVVAR_PREFIX_FOR_DYNACONF']
        )  # SPP_ENV

        dconf.setdefault(
            'ENV_FOR_DYNACONF',
            os.environ.get(env_prefix, 'DEV').upper()
        )

        dconf['SETTINGS_FILE_FOR_DYNACONF'] = os.path.join(settings_location,
                                                           settings_file)
        dconf['ROOT_PATH_FOR_DYNACONF'] = settings_location

        super().__init__(**dconf)




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['useismic',
 'useismic.clients',
 'useismic.clients.old_api_client',
 'useismic.core',
 'useismic.processors',
 'useismic.settings']

package_data = \
{'': ['*']}

install_requires = \
['dynaconf>=3.1.4,<4.0.0', 'uquake>=0.4.31,<0.5.0']

setup_kwargs = {
    'name': 'useismic',
    'version': '0.1.6',
    'description': '',
    'long_description': None,
    'author': 'jpmercier',
    'author_email': 'jpmercier01@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
